------------------------------------------------------------------------
-- PROJECT          :   MONSOON
-- BOARD            :   IR CLOCK&BIAS BOARD
-- DEVICE ID        :   VIRTEX FPGA (U116)
-- PARENT MODULE    :   TOP LEVEL
-- MODULE           :   OTA ROW / COLUMN SELECT MODULE
-- VERSION          :   4.91 
------------------------------------------------------------------------
-- FILE NAME        :   OT_RowColSelect.vhd
-- AUTHOR           :   Peter Moore
-- DATE             :   29 JANUARY 2006
-- COMPANY          :   NOAO
-- PLATFORM         :   Xilinx ISE Webpack 7.1
------------------------------------------------------------------------
-- DESCRIPTION:   
-- THIS MODULE CONTROLS THE MODE OF OTA CELLS BY USING THE ROW AND COLUMN 
-- SELECT AND D0,D1,D2 SIGNALS ASSIGNED TO THE CLOCK PORT BITS.
-- THERE ARE 8 MODES AVAILABLE THAT ARE SET BY WRITING A CODE TO THE MODE 
-- REGISTER:
-- CODE      MODE       DESCRIPTION
--   0    MOD_NOP       DO NOTHING
--   1    MOD_DISABLE   USE THE DISABLE MASK BIT REGISTERS TO SET THE 
--                      CORRESPONDING CELLS TO DISABLE MODE.
--   2    MOD_GS_SLCT   USE THE GUIDE MASK BIT REGISTERS TO SET THE 
--                      CORRESPONDING CELLS TO ACTIVE WITH VIDEO MODE
--   3    MOD_ROW_SLCT  DISABLE CURRENT ROW, INCREMENT ROW POINTER, SET
--                      ROW CELLS NOT DISABLED TO ACTIVE WITH VIDEO MODE
--   4    MOD_ROW_RST   DISABLE CURRENT ROW, SET ROW POINTER TO FIRST ROW
--   5    MOD_SHIFT     SET ALL CELLS THAT ARE NOT SET AS GUIDE CELLS TO
--                      ACTIVE WITH VIDEO OFF MODE
--   6    MOD_STANDBY   SET ALL CELLS THAT ARE NOT DISABLED TO STANDBY MODE
--   7    MOD_ACTIVE    SET ALL CELLS THAT ARE NOT DISABLED TO ACTIVE WITH
--                      VIDEO OFF MODE   
--
-- THERE ARE TWO SETS OF MASK REGISTERS, EACH SET IS 8 REGISTERS OF 8 BITS
-- EACH. EACH REGISTER IN A SET CORRESPONDS TO ONE ROW OF AN OTA DEVICE. EACH
-- BIT IN THAT REGISTER CORRESPONDS TO A CELL WITHIN THAT ROW. ONE SET (OF 8
-- REGISTERS) IS USED TO HOLD THE BIT PATTERNS THAT DESIGNATE CELLS THAT ARE 
-- DISABLED. THE OTHER SET DESIGNATES CELLS TOP BE USED FOR GUIDE OPERATIONS.   
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES:
--
------------------------------------------------------------------------
-- REVISION HISTORY:
-- DATE       AUTHOR  VERSION   MODIFICATIONS
--	01/21/2006   PCM     4.90    Module created as version 4.90
--  01/29/2006   PCM     4.91    CORRECTED PROBLEM WITH MISSING CLOCK 
--                               OVERLAP ON LAST COLUMN CLOCK. ADDED STATE
--                               TO MAINTAIN D2 STATE AFTER LAST CELL ACCESS 
------------------------------------------------------------------------
-- PORT DEFINITIONS:
--
-- SYSCLK         : SYSTEM CLOCK AT 40 MHZ
-- RESET          : SYSTEM RESET STROBE
-- WRITE16        : WRITE STROBE - ACTIVE HIGH
-- DATA_IN        : SEQUENCER DATA BUS
-- ADDR_IN        : SEQUENCER BUS ADDRESS BUS. MASK REGISTER POINTER
-- CS_DISABLE     : SELECT THE DISABLE MASK REGISTER SET	1 OF 8
-- CS_GUIDE       : SELECT THE GUIDE MASK REGISTER SET 1 OF 8
-- CS_MODE        : SELECT THE MODE REGISTER - WRITING TO THIS REGISTER INITAIATES A MODE CHANGE
-- GUIDE_DATA_OUT : GUIDE MASK REGISTER DATA ECHO
-- DISBALE_DATA_OUT : DISABLE MASK REGISTER DATA ECHO
-- MODE_DATA_OUT  : MODE REGISTER DATA ECHO
-- DOP_DATA_OUT   : DATA SENT TO THE CLOCK PORT VIA THE SEQ CONTROL MODULE
-- DOP_STROBE     : DATA READY STROBE TO INDICATE A WRITE REQUEST TO THE SEQUENCER CONTROL MODULE
-- DOP_ACK        : ACKNOWLEDGEMENT FROM THE SEQ CONTROL MODULE THAT THE VECTOR WAS WRITTEN
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity OT_RowColSelect is
    Port ( SYSCLK         : in  std_logic;
           RESET          : in  std_logic;
           WRITE16        : in  std_logic;
           WRITE32        : in  std_logic;
           DATA_IN        : in  std_logic_vector(7 downto 0);
           ADDR_IN        : in  std_logic_vector(2 downto 0);
           CS_DISABLE     : in  std_logic;
           CS_GUIDE       : in  std_logic;
           CS_MODE        : in  std_logic;
           GUIDE_DATA_OUT : out std_logic_vector(7 downto 0);
           DISABLE_DATA_OUT : out std_logic_vector(7 downto 0);
           MODE_DATA_OUT  : out std_logic_vector(7 downto 0);
           DOP_DATA_OUT   : out std_logic_vector(18 downto 0);
           DOP_STROBE     : out std_logic;
           DOP_ACK        : in  std_logic;
           CLK_DIVIDER    : in  std_logic_vector(7 downto 0)
			 );
end OT_RowColSelect;

architecture Behavioral of OT_RowColSelect is
    constant MOD_NOP      : std_logic_vector(2 downto 0) := "000";
    constant MOD_DISABLE  : std_logic_vector(2 downto 0) := "001";      -- Set all cells to disabled mode
    constant MOD_GS_SLCT  : std_logic_vector(2 downto 0) := "010";      -- Set guide cells to active w/video, science cells to standby
    constant MOD_ROW_SLCT : std_logic_vector(2 downto 0) := "011";      -- set all cells in current row to standby, increment row, set all cells in new row to active w/video (see note). 
    constant MOD_ROW_RST  : std_logic_vector(2 downto 0) := "100";      -- set all cells to standby, set row counter to 0.
    constant MOD_SHIFT    : std_logic_vector(2 downto 0) := "101";      -- set all guide cells to standby, set all science cells to active without video. 
    constant MOD_STANDBY  : std_logic_vector(2 downto 0) := "110";      -- set all cells to standby mode.
    constant MOD_ACTIVE   : std_logic_vector(2 downto 0) := "111";      -- set all cells to active without video.
                                                                        -- Note: For mode MOD_ROW_SLCT, if the previous mode set was a MOD_ROW_RST cmd, then just set current row to active w/video.

    constant CELL_DISABLE : std_logic_vector(2 downto 0) := not "001";   -- Dop signals are inverted. Make these 
    constant CELL_ACTIVE  : std_logic_vector(2 downto 0) := not "000";   -- codes high true when sent to detector
    constant CELL_VIDEO   : std_logic_vector(2 downto 0) := not "100";
    constant CELL_STANDBY : std_logic_vector(2 downto 0) := not "010";

-- Mask ram signals --
    type mask_ram_type is array (7 downto 0) of std_logic_vector (7 downto 0); 
    signal GuideStore   : mask_ram_type; -- Ram to store Guide mask
    signal DisableStore : mask_ram_type; -- Ram to store Disable mask
    signal GuideRamAddr   : std_logic_vector(2 downto 0);
    signal DisableRamAddr : std_logic_vector(2 downto 0);
    signal GuideRamData   : std_logic_vector(7 downto 0);
    signal DisableRamData : std_logic_vector(7 downto 0);
    signal RowAddr        : std_logic_vector(2 downto 0);

-- Mode and Control signals --
    signal ModeRegister   : std_logic_vector(7 downto 0);
    signal ModeWrtTrigger : std_logic;
    signal GuideMaskData  : std_logic_vector(7 downto 0);
    signal DisableMaskData: std_logic_vector(7 downto 0);
    signal DisableMask    : std_logic_vector(7 downto 0);
    signal DisableFlag    : std_logic;
    signal GuideMask      : std_logic_vector(7 downto 0);
    signal GuideFlag      : std_logic;
    signal CleanFlag      : std_logic;
    signal GuideContext   : std_logic;

-- CmdProc signals --
    type CmdStateType is (CmdIdle, CmdState_1, CmdState_2, CmdState_3, CmdState_4, CmdState_5, CmdState_6, CmdState_7, CmdState_end);
    signal CmdState       : CmdStateType;
    signal cmd_data       : std_logic_vector(2 downto 0);
    signal row_reset_flag : std_logic;
    signal cmd_row        : std_logic_vector(7 downto 0);
    signal cmd_busy_flag  : std_logic;

-- RowScanProc signals --
    type RowStateType is (RowIdle, RowState_1, RowState_2, RowState_3);
    signal RowState       : RowStateType;
    signal row_trigger    : std_logic;
    signal row_proc_busy  : std_logic;
    signal row_mode, row_data : std_logic_vector(2 downto 0);

-- SetCellMode signals --
    type SetCellStateType is (SetIdle, SetState_1, SetState_2, SetState_3, SetState_4, SetState_5, SetState_6);
    signal SetCellState       : SetCellStateType;
    signal cell_row, cell_col : std_logic_vector(7 downto 0);
    signal set_row, set_col   : std_logic_vector(7 downto 0);
    signal cell_data, set_data: std_logic_vector(2 downto 0);
    signal setcell_trigger    : std_logic;
    signal setcell_busy       : std_logic;
    signal set_clk_enable     : std_logic;
    signal set_clk_tick       : std_logic;
    signal persist_data       : std_logic_vector(2 downto 0);
    signal DopData            : std_logic_vector(18 downto 0);

-- TimingGen process signals --
    signal TimeBaseCtr  : std_logic_vector(7 downto 0);
    signal time_gen_ctr : std_logic_vector(7 downto 0);
    signal time_gen_tgl : std_logic;
begin

---------------------------------------------
-- Mask registers write / read processes   --
---------------------------------------------
    GuideRamAddr <= ADDR_IN when (CS_GUIDE = '1' and cmd_busy_flag = '0') else RowAddr;
    DisableRamAddr <= ADDR_IN when (CS_DISABLE = '1' and cmd_busy_flag = '0') else RowAddr;

    GuideRamWrite : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (CS_GUIDE = '1' and WRITE16 = '1') then                            
                GuideStore(conv_integer(GuideRamAddr)) <= DATA_IN;
            end if;
        end if;
    end process;

    GuideRamData <= GuideStore(conv_integer(GuideRamAddr));
    GUIDE_DATA_OUT <= GuideRamData;

    DisableRamWrite : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (CS_DISABLE = '1' and WRITE16 = '1') then                            
                DisableStore(conv_integer(DisableRamAddr)) <= DATA_IN;
            end if;
        end if;
    end process;

    DisableRamData <= DisableStore(conv_integer(DisableRamAddr));
    DISABLE_DATA_OUT <= DisableRamData;

---------------------------------------------
-- Mode registers write / read processes   --
---------------------------------------------
    ModeRegWrite : process (SYSCLK, RESET)
    begin
        if(RESET = '1') then
            ModeRegister <= (others => '0');
            ModeWrtTrigger <= '0';
        elsif(rising_edge(SYSCLK)) then
            if (CS_MODE = '1' and WRITE32 = '1') then                            
                ModeRegister <= DATA_IN;
                ModeWrtTrigger <= '1';
            else
                ModeWrtTrigger <= '0';
            end if;
        end if;
    end process;

    MODE_DATA_OUT <= ModeRegister;

------------------------------------------------------------------------
-- Mode Command Process is triggered by a write to the Mode Register  --
-- The command is processed and different state machines fired off to --
-- complete the task. This is the top level state machine             --
------------------------------------------------------------------------

    ModeCmdProc : process(SYSCLK, RESET)
    begin
        if(RESET = '1') then
            RowAddr <= (others => '0');
            row_trigger <= '0';
            GuideContext <= '0';
            cmd_busy_flag <= '0'; 
            CmdState <= CmdIdle;
        elsif (rising_edge(SYSCLK)) then
            case CmdState is
                when CmdIdle =>
                    if(ModeWrtTrigger = '1' and cmd_busy_flag = '0') then
                        cmd_busy_flag <= '1'; 
                        case ModeRegister(2 downto 0) is
                            when MOD_NOP =>
                                CmdState <= CmdIdle;
                            when MOD_DISABLE =>
                                cmd_data <= CELL_DISABLE;
                                cmd_row  <= "11111111";
                                GuideContext <= '0';
                                GuideMaskData <= "00000000";
                                DisableMaskData <= "00000000";
                                row_trigger <= '1';
                                CmdState <= CmdState_end;         -- Go hit on the array, just once
                            when MOD_STANDBY =>
                                cmd_data <= CELL_STANDBY;
                                RowAddr <= (others => '0');
                                cmd_row  <= "00000001";
                                row_reset_flag <= '1';
                                GuideContext <= '0';
                                GuideMaskData <= "00000000";
                                CmdState <= CmdState_1;         -- Begin row sequence
                            when MOD_ACTIVE =>
                                cmd_data <= CELL_ACTIVE;
                                RowAddr <= (others => '0');
                                cmd_row  <= "00000001";
                                GuideContext <= '0';
                                GuideMaskData <= "00000000";
                                CmdState <= CmdState_1;         -- Begin row sequence
                            when MOD_GS_SLCT =>
                                cmd_data <= CELL_VIDEO;
                                RowAddr <= (others => '0');
                                cmd_row  <= "00000001";
                                GuideContext <= '1';
                                CmdState <= CmdState_3;         -- Begin row sequence
                            when MOD_SHIFT =>
                                cmd_data <= CELL_ACTIVE;
                                RowAddr <= (others => '0');
                                cmd_row  <= "00000001";
                                GuideContext <= '0';
                                CmdState <= CmdState_3;         -- Begin row sequence
                            when MOD_ROW_SLCT =>
                                if(row_reset_flag = '0') then
                                    CmdState <= CmdState_5;         -- Set the current row to standby first, then increment row, and set active.
                                else
                                    row_reset_flag <= '0';
                                    CmdState <= CmdState_7;         -- The detector is in standby, just activate the current row (row 0).
                                end if;
                            when MOD_ROW_RST =>
                                cmd_data <= CELL_STANDBY;       -- This is redundant .. same as MOD_STANDBY
                                RowAddr <= (others => '0');
                                cmd_row  <= "00000001";
                                row_reset_flag <= '1';
                                GuideContext <= '0';
                                GuideMaskData <= "00000000";
                                CmdState <= CmdState_1;         -- Begin row sequence
                            when others =>
                                CmdState <= CmdIdle;
                        end case;
                    else
                        cmd_busy_flag <= '0'; 
                        CmdState <= CmdIdle;
                    end if;
                when CmdState_1 =>
                    if(cmd_row = "00000000") then      -- Test for end of row loop
                        RowAddr <= (others => '0');
                        cmd_row <= "00000001";         -- Setup for next time
                        CmdState <= CmdIdle;
                    elsif(row_proc_busy = '0') then
                        DisableMaskData <= DisableRamData; -- Grab the current row disable mask data
                        row_trigger <= '1';                -- Let the rowscan proc do its thing
                        CmdState <= CmdState_2;
                    else
                        CmdState <= CmdState_1;
                    end if;
                when CmdState_2 =>
                    if(row_proc_busy = '1') then           -- Wait for the rowscan proc to catch up
                        row_trigger <= '0';
                        RowAddr <= RowAddr + 1;               -- Increment the mask ram pointers
                        cmd_row <= cmd_row(6 downto 0) & '0'; -- shift the row pointer bit. 
                        CmdState <= CmdState_1;
                    else
                        CmdState <= CmdState_2;
                    end if;
                when CmdState_3 =>
                    if(cmd_row = "00000000") then      -- Test for end of row loop
                        RowAddr <= (others => '0');
                        cmd_row <= "00000001";         -- Setup for next time
                        CmdState <= CmdIdle;
                    elsif(row_proc_busy = '0') then
                        DisableMaskData <= DisableRamData; -- Grab the current row disable mask data
                        GuideMaskData <= GuideRamData;     -- and row guide mask data
                        row_trigger <= '1';                -- Let the rowscan proc do its thing
                        CmdState <= CmdState_4;
                    else
                        CmdState <= CmdState_3;
                    end if;
                when CmdState_4 =>
                    if(row_proc_busy = '1') then           -- Wait for the rowscan proc to catch up
                        row_trigger <= '0';
                        RowAddr <= RowAddr + 1;               -- Increment the mask ram pointers
                        cmd_row <= cmd_row(6 downto 0) & '0'; -- shift the row pointer bit. 
                        CmdState <= CmdState_2;
                    else
                        CmdState <= CmdState_4;
                    end if;
                when CmdState_5 =>                      -- Disable the current row
                    if(row_proc_busy = '0') then
                        cmd_data <= CELL_STANDBY;
                        GuideContext <= '0';
                        DisableMaskData <= DisableRamData; -- Grab the current row disable mask data
                        GuideMaskData <= GuideRamData;     -- and row guide mask data
                        row_trigger <= '1';                -- Let the rowscan proc do its thing
                        CmdState <= CmdState_6;
                    else
                        CmdState <= CmdState_6;
                    end if;
                when CmdState_6 =>
                    if(row_proc_busy = '1') then           -- Wait for the rowscan proc to catch up
                        row_trigger <= '0';
                        RowAddr <= RowAddr + 1;               -- Increment the mask ram pointers
                        cmd_row <= cmd_row(6 downto 0) & '0'; -- shift the row pointer bit. 
                        CmdState <= CmdState_7;
                    else
                        CmdState <= CmdState_6;
                    end if;
                when CmdState_7 =>                      -- Disable the current row
                    if(row_proc_busy = '0') then
                        cmd_data <= CELL_VIDEO;
                        GuideContext <= '0';
                        DisableMaskData <= DisableRamData; -- Grab the current row disable mask data
                        GuideMaskData <= GuideRamData;     -- and row guide mask data
                        row_trigger <= '1';                -- Let the rowscan proc do its thing
                        CmdState <= CmdState_end;
                    else
                        CmdState <= CmdState_7;
                    end if;
               when CmdState_end =>
                    if(row_proc_busy = '1') then
                        row_trigger <= '0';
                        CmdState <= CmdIdle;
                    else
                        CmdState <= CmdState_end;
                    end if;
                when Others =>
                    CmdState <= CmdIdle;
            end case;
        end if;
    end process;

---------------------------------------------------------
-- RowScanProc: Triggered by the CmdProc; Scans across --
-- a row of cells setting the required mode            --
---------------------------------------------------------
    RowScanProc : process(SYSCLK, RESET)
    begin
        if(RESET = '1') then
            cell_col <= "11111111"; 
            setcell_trigger <= '0';
            row_proc_busy <= '0';
        elsif(rising_edge(SYSCLK)) then
            case RowState is
                when RowIdle =>
                    if(row_trigger = '1' and setcell_busy = '0' and row_proc_busy = '0') then
                        row_proc_busy <= '1';
                        cell_row <= cmd_row;            -- capture the current row value
                        persist_data <= cmd_data;       -- This is to allow D2 to be active AFTER mode setting if required
                        if(CleanFlag = '1') then        -- Test to see if we can set the complete row  
                            cell_col  <= "11111111";    -- in one step i.e no guides nor disables in the row.
                            cell_data <= cmd_data;
                            setcell_trigger <= '1';
                            RowState <= RowState_1;
                        else
                            cell_col <= "00000001";     -- If there is a disabled or guide cell then we need to
                            row_mode <= cmd_data;
                            RowState <= RowState_2;     -- scan through individual cells
                        end if;
                    else
                        row_proc_busy <= '0';
                        RowState <= RowIdle;
                    end if;
                when RowState_1 =>                      -- Just wait for the completion of the row mode set
                    if(setcell_busy = '1') then
                        setcell_trigger <= '0';
                        cell_col <= "11111111";         -- Ready for next test case. 
                        RowState <= RowIdle;
                    else
                        RowState <= RowState_1;
                    end if;
                when RowState_2 =>
                    if(cell_col = "00000000") then      -- Test for end of column loop
                        cell_col <= "11111111";         -- Ready for next test case. 
                        RowState <= RowIdle;
                    else
                        cell_data <= row_data;          -- Recover the decoded mode for this cell
                        if(setcell_busy = '0') then
                            setcell_trigger <= '1';
                            RowState <= RowState_3;
                        else
                            RowState <= RowState_2;
                        end if;
                    end if;
                when RowState_3 =>
                    if(setcell_busy = '1') then
                        setcell_trigger <= '0';
                        cell_col <= cell_col(6 downto 0) & '0'; -- shift the column pointer bit. 
                        RowState <= RowState_2;
                    else
                        RowState <= RowState_3;
                    end if;
                when Others =>
                    RowState <= RowIdle;
            end case;
        end if;
    end process;

---------------------------------------------------------------
-- Cell mode decoder: Takes the commanded mode word and      --
-- resolves this into a 3 bit pattern  depending on the      --
-- mode required and the cell mode set in the mask registers --
---------------------------------------------------------------
    DisableMask <= DisableMaskData and cell_col;
    DisableFlag <= DisableMask(7) or DisableMask(6) or DisableMask(5) or DisableMask(4)
                or DisableMask(3) or DisableMask(2) or DisableMask(1) or DisableMask(0);

    GuideMask <= GuideMaskData and cell_col;
    GuideFlag <= GuideMask(7) or GuideMask(6) or GuideMask(5) or GuideMask(4)
                or GuideMask(3) or GuideMask(2) or GuideMask(1) or GuideMask(0);

    row_data <= CELL_DISABLE when (DisableFlag = '1') else                                -- Always force disable if cell is marked as such
                row_mode     when (GuideFlag = '0' and GuideContext = '0') else           -- Set the commanded mode if this cell is in science mode and isn't a guide cell
                CELL_STANDBY when (GuideFlag = '0' and GuideContext = '1') else           -- Set disable mode if this is a guide cell and we're setting science modes
                CELL_STANDBY when (GuideFlag = '1' and GuideContext = '0') else           -- or if we're trying to set guide modes and this is a science cell
                row_mode;                                                                 -- Otherwise we must be setting a guide mode and this is a guide cell.

    CleanFlag <= not (DisableFlag or GuideFlag);

------------------------------------------------------------------------
--                Truth table for mode select                         --
------------------------------------------------------------------------
--  ROW_MODE   | DISABLE_FLAG | GUIDE_FLAG | GUIDE_CONTEXT | ROW_DATA
------------------------------------------------------------------------
--CELL_DISABLE |      0       |     0      |       0       | CELL_DISABLE
--             |      0       |     0      |       1       | CELL_DISABLE
--             |      0       |     1      |       0       | CELL_DISABLE
--             |      0       |     1      |       1       | CELL_DISABLE
--             |      1       |     X      |       X       | CELL_DISABLE
--CELL_ACTIVE  |      0       |     0      |       0       | CELL_ACTIVE
--             |      0       |     0      |       1       | CELL_STANDBY 
--             |      0       |     1      |       0       | CELL_STANDBY
--             |      0       |     1      |       1       | CELL_ACTIVE
--             |      1       |     X      |       X       | CELL_DISABLE
--CELL_VIDEO   |      0       |     0      |       0       | CELL_VIDEO 
--             |      0       |     0      |       1       | CELL_STANDBY
--             |      0       |     1      |       0       | CELL_STANDBY
--             |      0       |     1      |       1       | CELL_VIDEO 
--             |      1       |     X      |       X       | CELL_DISABLE
--CELL_STANDBY |      0       |     0      |       0       | CELL_STANDBY
--             |      0       |     0      |       1       | CELL_STANDBY
--             |      0       |     1      |       0       | CELL_STANDBY
--             |      0       |     1      |       1       | CELL_STANDBY
--             |      1       |     X      |       X       | CELL_DISABLE
------------------------------------------------------------------------

----------------------------------------------------------------
-- Set cell process: Takes a cell row / column address and    --
-- a mode value and sequences the hardware clocks to set the  --
-- required mode on the physical detector.                    --
---------------------------------------------------------------- 

    SetCellMode : process(SYSCLK, RESET)
    begin
        if(RESET = '1') then
            setcell_busy <= '0';
            set_clk_enable <= '0';
            DopData <= (others => '0');
            DOP_STROBE <= '0';
            SetCellState <= SetIdle;
        elsif (rising_edge(SYSCLK)) then
            case SetCellState is
                when SetIdle =>
                    set_clk_enable <= '0';
                    if(setcell_trigger = '1') then
                        set_data <= cell_data;      -- Latch in the task parameters locally
                        set_col <= cell_col;
                        set_row <= cell_row;
                        setcell_busy <= '1';        -- tell the boss that we're busy
                        SetCellState <= SetState_1; -- Go start the state progression
                    else
                        setcell_busy <= '0';
                        SetCellState <= SetIdle;
                    end if;
                when SetState_1 =>
                    DopData <= set_row & set_data & x"00";     -- Strobe row with data and col low
                    DOP_STROBE <= '1';
                    set_clk_enable <= '1';
                    SetCellState <= SetState_2;
                when SetState_2 =>
                    if(DOP_ACK = '1') then
                        DOP_STROBE <= '0';
                    end if;
                    if(set_clk_tick = '1') then
                        DopData <= set_row & set_data & set_col;  -- Strobe row and col with data
                        DOP_STROBE <= '1';
                        SetCellState <= SetState_3;
                    else
                        SetCellState <= SetState_2;
                    end if;
                when SetState_3 =>
                    if(DOP_ACK = '1') then
                        DOP_STROBE <= '0';
                    end if;
                    if(set_clk_tick = '1') then
                        DopData <= x"00" & set_data & set_col;  -- Release row and keep col strobe on with data
                        DOP_STROBE <= '1';
                        SetCellState <= SetState_4;
                    else
                        SetCellState <= SetState_3;
                    end if;
                when SetState_4 =>
                    if(DOP_ACK = '1') then
                        DOP_STROBE <= '0';
                    end if;
                    if(set_clk_tick = '1') then
                        DopData <= x"00" & set_data & x"00";  -- Release row and col strobes on with data
                        DOP_STROBE <= '1';
                        SetCellState <= SetState_5;
                    else
                        SetCellState <= SetState_4;
                    end if;
                when SetState_5 =>
                    if(DOP_ACK = '1') then
                        DOP_STROBE <= '0';
                        SetCellState <= SetState_5;
                    elsif(set_clk_tick = '1') then                -- Wait for the correct clock overlap
                        DopData <= x"00" & persist_data & x"00";  -- Make sure that the data lines reflect the original mode request
                        DOP_STROBE <= '1';
                        SetCellState <= SetState_6;
                    else
                        SetCellState <= SetState_5;
                    end if;
                when SetState_6 =>
                    if(DOP_ACK = '1') then
                        DOP_STROBE <= '0';
                        SetCellState <= SetIdle;
                    else
                        SetCellState <= SetState_6;
                    end if;
                when Others =>
                    SetCellState <= SetIdle;
            end case;
        end if;           
    end process;

    DOP_DATA_OUT <= not DopData;
--------------------------------------------------
-- Timing generator for the setCellMode process --
--------------------------------------------------
    TimeGenProc : process(SYSCLK, set_clk_enable )
    begin
        if(set_clk_enable = '0') then
            set_clk_tick <= '0';
            time_gen_tgl <= '1';
            time_gen_ctr <= (others => '0');
            TimeBaseCtr <= (others => '0');
        elsif(rising_edge(SYSCLK)) then
            if(TimeBaseCtr(4) = '1') then          -- Prescaler .. Divide system clock by 16
                TimeBaseCtr <= (others => '0');
                if(time_gen_ctr > CLK_DIVIDER) then
                    time_gen_ctr <= (others => '0');
                    set_clk_tick <= '1';
                else
                    time_gen_ctr <= time_gen_ctr +1;
                end if;    
            else
                set_clk_tick <= '0';
                TimeBaseCtr <= TimeBaseCtr + 1;
            end if;
        end if;
    end process;

end Behavioral;
