------------------------------------------------------------------------
-- PROJECT          :   MONSOON
-- BOARD            :   IR CLOCK&BIAS BOARD
-- DEVICE ID        :   VIRTEX FPGA (U116)
-- PARENT MODULE    :   TOP LEVEL
-- MODULE           :   OTA SHIFT LOGIC - EMULATED CCD ACQUISITION BOARD PARALLEL CONTROL FUNCTIONS
-- VERSION          :   4.92 
------------------------------------------------------------------------
-- FILE NAME        :   OT_ShiftLogicV492.vhd
-- AUTHOR           :   Peter Moore
-- DATE             :   3rd December 2006
-- COMPANY          :   NOAO
-- PLATFORM         :   Xilinx ISE Webpack 7.1
------------------------------------------------------------------------
-- DESCRIPTION:   
-- HAS SEPERATE CONTROL OVER FOUR GROUPS OF FOUR SIGNALS EACH. EACH GROUP
-- CORRESPONDS TO THE PARALLEL CLOCK RAILS OF A RESPECTIVE OTA. TRIGGERING
-- ONE OF THESE GROUPS SEQUENCES THE PARALLEL CLOCK LINES TO SHIFT
-- CHARGE IN EITHER A POSITIVE (AWAY FROM AMPLIFIER) OR A NEGATIVE (TOWARDS
-- THE AMPLIFIER) DIRECTION. THE THIRTY TWO BIT WORD WRITTEN TO THE TRIGGER
-- SPECIFIES THE DIRECTION AND PIXEL COUNT FOR THE X AND Y AXIS OF THE 
-- DETECTOR
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES:
-- OTA SHIFT TRIGGER REGISTER
-- WRITING A 32 BIT WORD TO THESE ADDRESSES CAUSES A SERIES OF PARALLEL CLOCK 
-- PATTERNS TO BE SEQUENCED SO THAT THE ACTIVE CELLS ON AN OTA WILL HAVE THEIR 
-- PIXEL CHARGE SHIFTED. THE 32 BIT WORD CONTAINS TWO SIGNED 16 BIT VALUES AS
-- PER BELOW:
------------------------------------------------------------------------------
-- |                 31:16              |                  15:0               |
-- |  SIGNED ROW SHIFT VALUE (Y SHIFT)  | SIGNED COLUMN SHIFT VALUE (X SHIFT) |
------------------------------------------------------------------------------
-- BITS 31:16  TWOS COMPLIMENT VALUE FOR ROW SHIFT (Y-SHIFT).
-- BITS  15:0  TWOS COMPLIMENT VALUE FOR COLUMN SHIFT (X-SHIFT). 
-- THE INITIAL AND FINAL CLOCK PATTERN PRESENT ON THE PARALLELS IS
-- P1 = HIGH  = DOP BIT 19
-- P2 = HIGH  = DOP BIT 20
-- P3 = LOW   = DOP BIT 21
-- P4 = LOW	  = DOP BIT 22
-- THE SELECTED CELLS ARE Y-SHIFTED FIRST, THEN X-SHIFTED. 
--
------------------------------------------------------------------------
-- REVISION HISTORY:
-- DATE       AUTHOR  VERSION   MODIFICATIONS
--	12/03/2006   PCM     4.920   Built from ccd Acq Board OT_RowColSelect_V494.vhd
------------------------------------------------------------------------
-- PORT DEFINITIONS:
--
-- SYSCLK         : SYSTEM CLOCK AT 40 MHZ
-- RESET          : SYSTEM RESET STROBE
-- DATA_IN        : SEQUENCER DATA BUS
-- CS_SHIFT_A00   : WRITE STOBE TO TRIGGER SHIFT ON PARALLELS COUPLED TO OTA DEVICE 0
-- CS_SHIFT_A01   : WRITE STOBE TO TRIGGER SHIFT ON PARALLELS COUPLED TO OTA DEVICE 1
-- CS_SHIFT_A10   : WRITE STOBE TO TRIGGER SHIFT ON PARALLELS COUPLED TO OTA DEVICE 2
-- CS_SHIFT_A11   : WRITE STOBE TO TRIGGER SHIFT ON PARALLELS COUPLED TO OTA DEVICE 3
-- DOP_DATA_OUT   : DATA SENT TO THE CLOCK PORT VIA THE SEQ CONTROL MODULE
-- DOP_STROBE     : DATA READY STROBE TO INDICATE A WRITE REQUEST TO THE SEQUENCER CONTROL MODULE
-- DOP_ACK        : ACKNOWLEDGEMENT FROM THE SEQ CONTROL MODULE THAT THE VECTOR WAS WRITTEN
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity OT_ShiftLogicV492 is
    Port ( SYSCLK         : in  std_logic;
           RESET          : in  std_logic;
           DATA_IN        : in  std_logic_vector(31 downto 0);
           CS_SHIFT_A00   : in  std_logic;
           CS_SHIFT_A01   : in  std_logic;
           CS_SHIFT_A10   : in  std_logic;
           CS_SHIFT_A11   : in  std_logic;
           DOP_DATA_OUT   : out std_logic_vector(15 downto 0);
           DOP_STROBE     : out std_logic;
           DOP_ACK        : in  std_logic;
           CLK_DIVIDER    : in  std_logic_vector(7 downto 0);
           SHIFT_BUSY     : out std_logic
	      );
end OT_ShiftLogicV492;

architecture Behavioral of OT_ShiftLogicV492 is

    signal Shift_trigger      : std_logic;
    signal ShiftPattern       : std_logic_vector(3 downto 0);
    signal ShiftClk0, ShiftClk1, ShiftClk2, ShiftClk3 : std_logic_vector(3 downto 0);
    signal Idx                : std_logic_vector(1 downto 0);
    signal Busy, LoopBusy     : std_logic;

-- RunShift signals
    type ShiftStateType is (Shift_Idle, Shift_1, Shift_2, Shift_3);
    type ShiftLoopStateType is (ShiftLoop_Idle, ShiftLoop_1, ShiftLoop_2, ShiftLoop_3, ShiftLoop_4, ShiftLoop_5);
    signal ShiftState         : ShiftStateType;
    signal ShiftLoopState     : ShiftLoopStateType;
    signal XCount, YCount     : std_logic_vector(15 downto 0);
    signal ShiftDopStrobe     : std_logic;
    signal shift_clk_enable   : std_logic;
    signal shift_clk_tick     : std_logic;
    signal ShiftX             : std_logic;
    signal XShiftPattern      : std_logic_vector(3 downto 0);
    signal YShiftPattern      : std_logic_vector(3 downto 0);
    signal shift_positive     : std_logic;
    signal ShiftBusy, ShiftEnd : std_logic;
    signal ShiftIndex         : std_logic_vector(3 downto 0);
    signal ShiftTrigger       : std_logic;
    signal ShiftIdxPtr        : std_logic_vector(3 downto 0); 
	signal ShiftCount         : std_logic_vector(3 downto 0);

-- TimingGen process signals --
    signal ShTimeBaseCtr : std_logic_vector(2 downto 0);
    signal ShTimeGenCtr  : std_logic_vector(7 downto 0);
begin

------------------------------------------------------------
-- ONLY ALLOW ONE SHIFT COMMAND TO BE PROCESSED AT A TIME --
------------------------------------------------------------
    Shift_trigger <= CS_SHIFT_A11 or CS_SHIFT_A10 or CS_SHIFT_A01 or CS_SHIFT_A00;
    Busy <= LoopBusy or ShiftBusy;
    SHIFT_BUSY <= Busy;

    ShiftTriggerProc : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK) and Busy = '0') then
            if (CS_SHIFT_A11 = '1') then
                Idx <= "11";
            elsif (CS_SHIFT_A10 = '1') then
                Idx <= "10";
            elsif (CS_SHIFT_A01 = '1') then
                Idx <= "01";
            elsif (CS_SHIFT_A00 = '1') then
                Idx <= "00";
            end if;
        end if;
    end process;

---------------------------------------------------------
-- ROW / COLUMN PIXEL SHIFTING CLOCK PATTERN GENERATOR --
---------------------------------------------------------
-- Loop counter for number of shifts
    ShiftLoopProcess : process(SYSCLK, RESET)
    begin
        if(RESET = '1') then
            ShiftTrigger <= '0';
            ShiftLoopState <= ShiftLoop_Idle;
            shift_positive <= '0';
            LoopBusy <= '0';
        elsif (rising_edge(SYSCLK)) then
            case ShiftLoopState is
                when ShiftLoop_Idle =>
                    if(LoopBusy = '0' and Shift_trigger = '1') then
                        XCount <= DATA_IN(15 downto 0);
                        YCount <= DATA_IN(31 downto 16);
                        ShiftX <= '0';
                        ShiftEnd <= '0';
                        shift_positive <= not DATA_IN(31);
                        LoopBusy <= '1';
                        ShiftLoopState <= ShiftLoop_1;
                    else
                        LoopBusy <= '0';
                        ShiftLoopState <= ShiftLoop_Idle;
                    end if;
                when ShiftLoop_1 =>
                    if (YCount = x"0000") then
                        ShiftEnd <= '1';
                        if (XCount = x"0000") then
                            ShiftLoopState <= ShiftLoop_Idle;
                        else
                            ShiftLoopState <= ShiftLoop_3;
                        end if;
                    else
                        if (ShiftBusy = '0') then
                            ShiftTrigger <= '1';
                            ShiftLoopState <= ShiftLoop_2;
                        else
                            ShiftLoopState <= ShiftLoop_1;
                        end if;
                    end if;
                when ShiftLoop_2 =>
                    if (ShiftBusy = '1') then
                        ShiftTrigger <= '0';
                        if (shift_positive = '1') then
                            YCount <= YCount - 1;
                        else
                            YCount <= YCount + 1;
                        end if;
                        ShiftLoopState <= ShiftLoop_1;
                    else
                        ShiftLoopState <= ShiftLoop_2;
                    end if;
                when ShiftLoop_3 =>
                    if (ShiftBusy = '0') then
                        shift_positive <= not XCount(15);
                        ShiftX <= '0';
                        ShiftEnd <= '0';
                        ShiftLoopState <= ShiftLoop_4;
                    else
                        ShiftLoopState <= ShiftLoop_3;
                    end if;
                when ShiftLoop_4 =>
                    if (XCount = x"0000") then
                        ShiftEnd <= '1';
                        ShiftLoopState <= ShiftLoop_Idle;
                    else
                        if (ShiftBusy = '0') then
                        	 ShiftX <= '1';
                            ShiftTrigger <= '1';
                            ShiftLoopState <= ShiftLoop_5;
                        else
                            ShiftLoopState <= ShiftLoop_4;
                        end if;
                    end if;
                when ShiftLoop_5 =>
                    if (ShiftBusy = '1') then
                        ShiftTrigger <= '0';
                        if (shift_positive = '1') then
                            XCount <= XCount - 1;
                        else
                            XCount <= XCount + 1;
                        end if;
                        ShiftLoopState <= ShiftLoop_4;
                    else
                        ShiftLoopState <= ShiftLoop_5;
                    end if;
            end case;
        end if;
    end process;

--    YShiftPattern <=    x"3" when ShiftIndex = x"0" else
--                        x"2" when ShiftIndex = x"1" else
--                        x"6" when ShiftIndex = x"2" else   ------
--                        x"2" when ShiftIndex = x"3" else   --    |
--                        x"3" when ShiftIndex = x"4" else   --    |
--                        x"1" when ShiftIndex = x"5" else   --    | Main shift loop
--                        x"5" when ShiftIndex = x"6" else   --    |
--                        x"4" when ShiftIndex = x"7" else   --    |
--                        x"6" when ShiftIndex = x"8" else   ------
--                        x"2" when ShiftIndex = x"9" else
--                        x"3" when ShiftIndex = x"A" else
--                        x"3";

--  XShiftPattern <=    x"3" when ShiftIndex = x"0" else
--                      x"2" when ShiftIndex = x"1" else
--                      x"6" when ShiftIndex = x"2" else   ------
--                      x"2" when ShiftIndex = x"3" else   --    |
--                      x"A" when ShiftIndex = x"4" else   --    |
--                      x"8" when ShiftIndex = x"5" else   --    | Main shift loop
--                      x"C" when ShiftIndex = x"6" else   --    |
--                      x"4" when ShiftIndex = x"7" else   --    |
--                      x"6" when ShiftIndex = x"8" else   ------
--                      x"2" when ShiftIndex = x"9" else
--                      x"3" when ShiftIndex = x"A" else
--                      x"3";


    XShiftPattern <=    x"2" when ShiftIndex = x"0" else
                        x"6" when ShiftIndex = x"1" else
                        x"4" when ShiftIndex = x"2" else   ------
                        x"C" when ShiftIndex = x"3" else   --    |
                        x"8" when ShiftIndex = x"4" else   --    |
                        x"A" when ShiftIndex = x"5" else   --    | Main shift loop
                        x"2" when ShiftIndex = x"6" else   --    |
                        x"2";

    YShiftPattern <=    x"2" when ShiftIndex = x"0" else  ------
                        x"6" when ShiftIndex = x"1" else  --    |
                        x"4" when ShiftIndex = x"2" else  --    | 
                        x"5" when ShiftIndex = x"3" else  --    | Main shift loop 
                        x"1" when ShiftIndex = x"4" else  --    | 
                        x"3" when ShiftIndex = x"5" else  ------ 
                        x"2" when ShiftIndex = x"6" else
                        x"2";


     ShiftIdxPtr <= x"0" when (shift_positive = '1') else x"6";
     ShiftPattern <= XShiftPattern when ShiftX = '1' else YShiftPattern;

     ShiftClk0 <= ShiftPattern when Idx = "00" else ShiftClk0;
     ShiftClk1 <= ShiftPattern when Idx = "01" else ShiftClk1;
     ShiftClk2 <= ShiftPattern when Idx = "10" else ShiftClk2;
     ShiftClk3 <= ShiftPattern when Idx = "11" else ShiftClk3;

--
-- state machine that sequences the actual clock waveforms.
--
    RunShiftProcess : process(SYSCLK, RESET)
    begin
        if(RESET = '1') then
            ShiftBusy <= '0';
            shift_clk_enable <= '0';
            ShiftDopStrobe <= '0';
            ShiftState <= Shift_Idle;
            ShiftIndex <= (others => '0');
        elsif (rising_edge(SYSCLK)) then
            case ShiftState is
                when Shift_Idle =>
                    if(ShiftTrigger = '1') then
                        ShiftBusy <= '1';              -- tell the boss that we're busy
                        shift_clk_enable <= '1';       -- start the overlap clock counter
                        ShiftCount <= x"6";
                        ShiftIndex <= ShiftIdxPtr; -- Go start the state progression presuming that charge is under parallels 2&3
                        ShiftState <= Shift_1;
                    else
                        shift_clk_enable <= '0';
                        ShiftBusy <= '0';
                        ShiftState <= Shift_Idle;
                    end if;
                when Shift_1 =>             -- Issue a dop strobe if we need to move charge.
                    if (ShiftCount = x"0") then
                        if (ShiftEnd = '1') then
                            ShiftDopStrobe <= '1';
                            ShiftState <= Shift_3;
                        else
                            ShiftState <= Shift_Idle;
                        end if;
                    else
                        ShiftDopStrobe <= '1';
                        ShiftCount <= ShiftCount - 1;
                        ShiftState <= Shift_2;
                    end if;
                when Shift_2 =>             -- Wait for the DOP acknowledgement and another clock tick.
                    if(DOP_ACK = '1') then
                        ShiftDopStrobe <= '0';
                    end if;
                    if(shift_clk_tick = '1') then
                        if (shift_positive = '1') then
                            ShiftIndex <= ShiftIndex + 1;
                        else
                            ShiftIndex <= ShiftIndex - 1;
                        end if;
                        ShiftState <= Shift_1;
                    else
                        ShiftState <= Shift_2;
                    end if;
                when Shift_3 =>             -- Put the charge back under the right phase
                    if(DOP_ACK = '1') then
                        ShiftDopStrobe <= '0';
                    end if;
                    if(shift_clk_tick = '1') then
                        ShiftState <= Shift_Idle;
                    end if;
            end case;
        end if;
    end process;

    DOP_DATA_OUT <= ShiftClk3 & ShiftClk2 & ShiftClk1 & ShiftClk0;
    DOP_STROBE <= ShiftDopStrobe;

--------------------------------------------------
-- Timing generator for the shift process       --
--------------------------------------------------

    ShiftTimeGenProc : process(SYSCLK, shift_clk_enable )
    begin
        if(shift_clk_enable = '0') then
            shift_clk_tick <= '0';
            ShTimeGenCtr <= x"01";
            ShTimeBaseCtr <= "001";
        elsif(rising_edge(SYSCLK)) then
            if(ShTimeBaseCtr(2) = '1') then          -- Prescaler .. Divide system clock by 4 = 100ns
                ShTimeBaseCtr <= "001";
                if(ShTimeGenCtr >= CLK_DIVIDER) then
                    ShTimeGenCtr <= x"01";
                    shift_clk_tick <= '1';
                else
                    ShTimeGenCtr <= ShTimeGenCtr +1;
                end if;    
            else
                shift_clk_tick <= '0';
                ShTimeBaseCtr <= ShTimeBaseCtr + 1;
            end if;
        end if;
    end process;

end Behavioral;
