------------------------------------------------------------------------
-- PROJECT      :   MONSOON
-- BOARD        :   GENERIC
-- DEVICE ID    :   ANY 300e SERIES FPGA
-- PARENT MODULE:   ANY TOP LEVEL DESIGN  
-- MODULE       :   SEQUENCER BUS INTERFACE
-- VERSION      :   7.9/CB 
------------------------------------------------------------------------
-- FILE NAME    :   CB_SEQ_INTERFACE_709.VHD
-- AUTHOR       :   PETER MOORE
-- DATE         :   04/03/2004
-- COMPANY      :   NOAO
-- PLATFORM     :   GENERIC Xilinx ISE
------------------------------------------------------------------------
-- DESCRIPTION  :   
-- IMPLEMENTS A STANDARD INTERFACE BETWEEN THE MONSOON BUS SYSTEM AND
-- ANY PERIPHERAL BOARD.
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- THIS IMPLEMENTATION CONFORMS TO MONSOON ICD 7.9
-- DOCUMENT NUMBER MNSN-xx-XX-xx
-- IT GENERALLY IS APPLICABLE TO ALL NOAO MONSOON PERIPHERAL BOARDS UNDER
-- REVISION '0'.
--
-- THIS MODULE CONTAINS THE FOLLOWING SUB ASSEMBLIES:
-- CLOCK                    : USED AND PASSED STRAIGHT THROUGH    
-- BOARD SELECT             : ENABLES THE MODE DECODE LOGIC
-- MODE                     : DECODES THE SEQUENCER MODE AND ASSERTS THE CORRECT SIGNALS
-- RESET                    : SYNCHRONOUSLY DECODED TO PROVIDE SOFT RESET AND HARD REBOOT SIGNALS
-- ADDRESS MUX              : SELECTS THE CORRECT ADDRESS DEPENDING ON WHETHER 32 OR 16 BIT MODE ASSERTED
-- SEQ BUS DATA MUX         : DEMULTIPLEXES THE DATA FROM THE BUS.
-- PIXEL BUS TRISTATES      : ENABLES THE PIXEL DATA BUS TO TRANSMIT DATA TO THE MCB
--
-- THE VISION IS THAT USING THIS TEMPLATE AS A STARTING POINT, ANY FUTURE
-- PERIPHERAL BOARD WILL BE ABLE TO PERSONALIZE THIS MODULE TO SUIT THEIR
-- PARTICULAR REQUIREMENTS. TO USE THIS TEMPLATE, IMPORT IT INTO YOUR DESIGN
-- AS A LOCAL PROJECT FILE AND HACK AWAY TO MAKE IT WORK FOR YOU.
------------------------------------------------------------------------
-- REVISION HISTORY :
--  DATE       AUTHOR VERSION  MODIFICATIONS
--  01/21/2003   PCM    0.1    ORIGINAL VERSION
--  04/02/2004   PCM    7.10   REMOVED SELF TEST LOGIC AND INTEGRATED 
--                             COMPLETE IFC INTO ONE FILE. 
--  04/03/2004   PCM    7.9    ADAPTED FOR REVISON 0 BOARDS I.E. ICD 7.9
--  May/03/2004  GRB   7.9/CB  Personalized for the Clock&Bias Board
------------------------------------------------------------------------
-- PORT DEFINITIONS :
-- SEQ_CLK  : SYSTEM CLOCK FROM THE BACKPLANE - ENABLED/DISABLED ON THE MCB.
-- SYSCLK   : LOCAL CLOCK OUT FOR ALL ON BOARD USAGE
-- BRDSEL_N : SEQUENCER BUS BOARD SELECT STROBE FROM MCB. LOW TRUE
-- BOARD_SEL: LOCAL INVERTED BOARD SELECT STROBE. 
-- BUSY     : LOCAL BUSY SIGNAL. NORMALLY TIED LOW.
-- BRDACK_N : LOCAL BOARD SELECT ACKNOWLEDGE OUTPUT STROBE TO MCB. CURRENTLY NOT USED BY MCB 
-- SEQ_MODE : SEQUENCER BUS MODE BITS FROM MCB.
-- RESET    : LOCAL BOARD RESET OUTPUT. TRUE FOR ONE CLOCK PERIOD FOR SOFT RESET ON BOARD.
-- REBOOT_N : LOCAL BOARD LATCHED REBOOT SIGNAL. LOW TRUE. NORMALLY ATTACHED TO PROGRAM_N FPGA SIG.
-- READ     : LOCAL DECODED CONTROL STROBE. READ SOMETHING.
-- WRITE    : LOCAL DECODED CONTROL STROBE. WRITE SOMETHING. 
-- WRT16    : LOCAL DECODED CONTROL STROBE. WRITE SOMETHING AS 16 BITS WIDE.
-- WRT32    : LOCAL DECODED CONTROL STROBE. WRITE SOMETHING AS 32 BITS WIDE.
-- DEV_ADDR : SEQUENCER BUS DEVICE ADDRESS FROM MCB. USED DURING RESET AND 32 BIT WRITES.
-- SEQ_DATA : SEQUENCER BUS DATA / ADDRESS BUS FROM MCB. 
-- ADDRESS  : LOCAL ADDRESS FOR USE ON BOARD 
-- DATA_OUT : LOCAL DATA FOR USE ON BOARD. ONLY LOWER 16 BITS VALID DURING A WRT16 STROBE.
-- DATA_IN  : LOCAL DATA BUS FROM BOARD TO MCB.
-- PIX_DATA : PIXEL DATA BUS TO MCB. TRISTATED UNLESS A READ IN PROGRESS.
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity seq_ifc_709 is
    Port ( 
            SEQ_CLK  : in  std_logic;
            SYSCLK   : out std_logic;
            BRDSEL_N : in  std_logic;
            BOARD_SEL: out std_logic;
            BUSY     : in  std_logic;
            BRDACK_N : out std_logic;
            SEQ_MODE : in  std_logic_vector(1 downto 0);
            RESET    : out std_logic;
            REBOOT_N : out std_logic;
            READ     : out std_logic;
            WRITE    : out std_logic;
            WRT16    : out std_logic;
            WRT32    : out std_logic;
            DEV_ADDR : in  std_logic_vector(2 downto 0);
            SEQ_DATA : in  std_logic_vector(31 downto 0);
            ADDRESS  : out std_logic_vector(15 downto 0);
            DATA_OUT : out std_logic_vector(31 downto 0);
            DATA_IN  : in  std_logic_vector(47 downto 0);
            PIX_DATA : out std_logic_vector(47 downto 0)
          );
end seq_ifc_709;

architecture Behavioral of seq_ifc_709 is
-- LOCAL MODE DECODED SIGNALS
    signal loc_read : std_logic;

-- RESET STATE LOGIC SIGNALS
    signal DevAddrZero : std_logic;      -- Used to determine if a reset state is legitimate.
    signal DevAddrOnes : std_logic;      -- Used to determine if reboot state is legitimate.
    signal loc_rst, loc_rbt : std_logic; -- For our internal use.

-- PIXEL BUS CONTROL SIGNAL
    signal loc_enbl : std_logic;

begin

------------------------------------------------------------------------
-- CLOCK SOURCE BEHAVIOUR
------------------------------------------------------------------------
    SYSCLK <= SEQ_CLK;

------------------------------------------------------------------------
-- BOARD SELECT BEHAVIOUR
------------------------------------------------------------------------
    BRDACK_N  <= not BUSY;
    BOARD_SEL <= not BRDSEL_N;

------------------------------------------------------------------------
-- MODE BEHAVIOUR
------------------------------------------------------------------------
    ModeDecode : process (SEQ_MODE, BRDSEL_N)
    begin
        if (BRDSEL_N = '0') then
            case SEQ_MODE is
                when "00" =>
                    loc_read  <= '0';
                    WRT16 <= '0';
                    WRT32 <= '0';
                    WRITE <= '0';
                when "01" =>
                    loc_read  <= '1';
                    WRT16 <= '0';
                    WRT32 <= '0';
                    WRITE <= '0';
                when "10" =>
                    loc_read  <= '0';
                    WRT16 <= '1';
                    WRT32 <= '0';
                    WRITE <= '1';
                when "11" =>
                    loc_read  <= '0';
                    WRT16 <= '0';
                    WRT32 <= '1';
                    WRITE <= '1';
                when others =>
                    null;
            end case;
        else
            loc_read  <= '0';
            WRT16 <= '0';
            WRT32 <= '0';
            WRITE <= '0';
        end if;
    end process;

    READ <= loc_read;

------------------------------------------------------------------------
-- RESET BEHAVIOUR
------------------------------------------------------------------------
    DevAddrZero <= not (BRDSEL_N or DEV_ADDR(2) or DEV_ADDR(1) or DEV_ADDR(0));
    DevAddrOnes <= (not BRDSEL_N) and DEV_ADDR(2) and DEV_ADDR(1) and DEV_ADDR(0);

    ResetState : process(SEQ_CLK)
    begin
        if(rising_edge(SEQ_CLK)) then
            if(SEQ_MODE = "00") then
                loc_rst <= DevAddrZero;
                loc_rbt <= DevAddrOnes or loc_rbt;
            else
                loc_rst <= '0';           -- Note that reboot is latched forever.
            end if;
        end if;
    end process;
    
    RESET  <= loc_rst;
    REBOOT_N <= not loc_rbt;           

------------------------------------------------------------------------
-- ADDRESS MUX BEHAVIOUR
------------------------------------------------------------------------
    ADDRESS <= ("0000000000000" & DEV_ADDR(2 downto 0)) when(SEQ_MODE = "11") else SEQ_DATA(31 downto 16) ;

------------------------------------------------------------------------
-- SEQ DATA MUX BEHAVIOUR
------------------------------------------------------------------------
    DATA_OUT <=  SEQ_DATA when(SEQ_MODE = "11") else ("0000000000000000" & SEQ_DATA(15 downto 0));

------------------------------------------------------------------------
-- PIXEL DATA BUS BEHAVIOUR
------------------------------------------------------------------------
    loc_enbl <= not BRDSEL_N and loc_read;
    PIX_DATA <= DATA_IN when(loc_enbl = '1') else (others => 'Z');

   
end Behavioral;
