------------------------------------------------------------------------
-- PROJECT          :   MONSOON
-- BOARD            :   IR CLOCK&BIAS BOARD
-- DEVICE ID        :   VIRTEX FPGA (U116)
-- PARENT MODULE    :   TOP LEVEL
-- MODULE           :   SERIAL CONFIGURATION CONTROL MODULE
-- VERSION          :   1.4 
------------------------------------------------------------------------
-- FILE NAME        :   cfg_control.vhd
-- AUTHOR           :   G. Rahmer
-- DATE             :   14 JANUARY 2003
-- COMPANY          :   NOAO
-- PLATFORM         :   Xilinx ISE Webpack 5.1
------------------------------------------------------------------------
-- DESCRIPTION:   
-- This module provides the control signals to operate the serial data 
-- stream going to the CFG CPLD.
--
-- The CFG_SEL bits are set according to the type of data, which is 
-- defined by the corresponding address:
--      00->DAC
--      01->ADC
--      10->OE
--      11->MUX
--
-- The WRITE_SD signal is used to trigger the serial data stream 
-- generation on cfg_ser_out, where the parallel data read from 
-- cfg_reg_file will be serialized and sent to the CFG CPLD.
--
-- The ADC_DATA_READY input signal is set by the cfg_ser_in module when
-- there is data ready to be written into cfg_reg_file (in the 
-- corresponding "ADC Data" memory location).
-- 
-- The CFGCLK which goes to the CFG CPLD is generated here from the
-- board clock (SYSCLK), and is turned on only when accessing the CPLD.
-- A backplane generated reset condition (at RST) is passed to 
-- cfg_reg_file through CFG_RST only when the register file is selected
-- (CS is on). 
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES:
-- A backplane "write" to an "ADC Data" memory address triggers an ADC 
-- access to the channel which corresponds to that memory location. The 
-- configuration info that has to be sent to the ADC must have been 
-- stored previously in the corresponding address of the "ADC Config" 
-- area of the memory. I DO NOT check that something appropriate was written.
--
-- In order for the CFG CPLD to get the proper ADC_SSTRB signal from the 
-- selected ADC, all ADC chips must be set to "external clock mode", 
-- otherwise the ADC_SSTRB line will not be high-Z when the chip is not selected.
-- (see corresponding red flag doc). A timeout of 8 CFGCLK cycles
-- is included in this module in case there is no ADC_SSTRB comming (which happens
-- normally 5 cycles after the end of the config program).
-- The timeout condition is assigned to the LSB of STATUS.
--
-- CFGCLK is 1/32 of the board clock (40 Mhz), i.e. 1.25 MHz, to make sure
-- that the slowest components (ADCs) are properly programmed. Max clock
-- for the ADCs is 2 MHz.
-- This is done with the "cfgclk_base" signal.
--
------------------------------------------------------------------------
-- REVISION HISTORY:
-- DATE       AUTHOR  VERSION   MODIFICATIONS
-- 07/Apr/03    GR      1.1     Basic code completed
-- 01/Jul/03    GR      1.2     Changed cfg clock divide ratio to 32, cleaned 
--                              up the code
-- 06/Jul/03    GR      1.21    Included a timeout condition for the ADC 
--                              readback, to prevent a lock-up situation
--                              (not used now)
--                              Added a delay in the cfg_reset state to allow
--                              external state machines to finish properly.
-- 01/Jul/03    GR      1.3     Modified STATUS port
-- 10/May/04    GR      1.4     Fixed ADC readback logic
--
------------------------------------------------------------------------
-- PORT DEFINITIONS:
--
-- SYSCLK       : board clock
-- RESET        : reset signal from the backplane interface
-- WR_BACKPLANE : write signal from the backplane interface (16-bit write)
-- CS_CFG       : chip select for the cfg_data reg file, from data_mux
-- CS_TEL       : chip select for the tel_data reg file, from data_mux
-- ADDR_BACKPLANE(7:0) : address from the backplane interface
-- DATA_BACKPLANE(15:0) : backplane data bus
-- CONTROL(15:0)   : control register word (not used)
-- ADC_DATA_READY  : input to indicate that ADC data is ready, from block
--                   "cfg_ser_in"
-- CFG_DONE     : input from "cfg_ser_out" to indicate that the tx of the 
--                serial stream has been completed.
-- ADDR_BACKPLANE_L(7:0) : backplane address latched with WR_BACKPLANE
-- DATA_BACKPLANE_L(7:0) : backplane data latched with WR_BACKPLANE
-- CFGCLK       : clock for cfg modules and the CFG CPLD (output)
-- WRITE_SD     : output signal to trigger the generation of the serial 
--                stream in block "cfg_ser_out".
-- RESET_CFG    : local reset for cfg_ser_out and cfg_ser_in
-- CFG_SEL(1:0) : function select for the CFG CPLD (output)
-- CFG_STATUS(7:0) : status word for the module
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity cfg_control is
    Port ( SYSCLK : in std_logic;
           RESET : in std_logic;
           WR_BACKPLANE : in std_logic;
           CS_CFG : in std_logic;
           CS_TEL : in std_logic;
           ADDR_BACKPLANE : in std_logic_vector(7 downto 0);
           DATA_BACKPLANE : in std_logic_vector(15 downto 0);
           ADC_DATA_READY : in std_logic;
           CFG_DONE : in std_logic;
           ADDR_BACKPLANE_L :out std_logic_vector(7 downto 0);
           DATA_BACKPLANE_L :out std_logic_vector(15 downto 0);
           CFGCLK : out std_logic;
           WRITE_SD : out std_logic;
           RESET_CFG : out std_logic;
           timeout : out std_logic;
           CFG_SEL : out std_logic_vector(1 downto 0);
           CFG_STATUS : out std_logic_vector(7 downto 0));
end cfg_control;

architecture Behavioral of cfg_control is
    
    signal wr_backplane_cfg : std_logic;

    signal divide : std_logic_vector(4 downto 0);
    signal cfgclk_base : std_logic;
    signal cfgclk_en : std_logic;

    signal adc_timeout_counter : std_logic_vector(9 downto 0);
    signal adc_timeout : std_logic;                             -- TRUE when adc_timeout_counter
                                                                -- reaches x"200" (eq. 32x16)

    signal reset_state_counter : std_logic_vector(5 downto 0);  -- delay for the cfg_reset state

    type statetype is (idle, cfg_active, adc_readback, cfg_reset);
    signal state : statetype;
    signal fsm_state : std_logic_vector(2 downto 0);
    signal cfg_function : std_logic_vector(1 downto 0);
    signal adc_cfg_data : std_logic;

begin
    
    wr_backplane_cfg <= WR_BACKPLANE and (CS_CFG or CS_TEL);    -- this module is selected
       
    backplane_AD_latch: process(SYSCLK, WR_BACKPLANE, ADDR_BACKPLANE, DATA_BACKPLANE)
    begin
        if rising_edge(SYSCLK) then
            if (wr_backplane_cfg = '1') then
                ADDR_BACKPLANE_L <= ADDR_BACKPLANE;
                DATA_BACKPLANE_L <= DATA_BACKPLANE;
            end if;
        end if;
    end process;

  
    cfg_function_latch: process(SYSCLK, WR_BACKPLANE, ADDR_BACKPLANE, CS_CFG, CS_TEL)
    begin
        if rising_edge(SYSCLK) then
            if (wr_backplane_cfg = '1') then
                if (CS_CFG = '1' and ADDR_BACKPLANE = x"6F") then       
                    cfg_function <= "11";                               -- MUX
                    adc_cfg_data <= '0';
                elsif (CS_CFG = '1' and ADDR_BACKPLANE = x"6E") then    
                    cfg_function <= "10";                               -- OE
                    adc_cfg_data <= '0';
                elsif (CS_TEL = '1') then                               
                    cfg_function <= "01";                               -- ADC cfg trigger
                    adc_cfg_data <= '0';
                elsif (CS_CFG = '1' and ADDR_BACKPLANE < x"70") then
                    cfg_function <= "00";                               -- DAC cfg
                    adc_cfg_data <= '0';
                else
                    adc_cfg_data <= '1';                                -- ADC cfg data
                    cfg_function <= "01";                               
                end if;
            end if;
        end if;
    end process;

    CFG_SEL <= cfg_function;
        
    clock_divider: process(SYSCLK)
    begin
        if rising_edge(SYSCLK) then
            if (RESET = '1') then
                divide <= "00000";
	        elsif (cfgclk_en = '1') then
                divide <= divide + '1';
            end if;
        end if;
    end process;

    cfgclk_base <= divide(4);             -- 1/32 of board clock
--  cfgclk_base <= divide(0);             -- 1/2 of board clock for simulation purposes

    CFGCLK <= cfgclk_base and cfgclk_en;

    adc_timeout <= adc_timeout_counter(9);  -- bit for a count of 32 times 16 (period of CFGCLK)

    TIMEOUT <= adc_timeout;

    CFG_STATUS <= fsm_state & "0000" & adc_timeout;

    Main_FSM: process(SYSCLK)
    begin
 	    if (rising_edge(SYSCLK)) then
        	if (RESET = '1') then
	    		state <= idle;
                cfgclk_en <= '0';
    		    WRITE_SD <= '0';
                RESET_CFG <= '1';
                adc_timeout_counter <= (others => '0');
                reset_state_counter <= (others => '0'); 
		    else
				case state is
				    when idle =>
                        fsm_state <= "100";
                        cfgclk_en <= '0';
                        reset_state_counter <= (others => '0'); 
                        WRITE_SD <= '0';
                        RESET_CFG <= '0';
						if (wr_backplane_cfg = '1' and adc_cfg_data = '0') then
						    state <= cfg_active;
						else
						    state <= idle;	 	
						end if;
                    when cfg_active =>
                        fsm_state <= "101";
                        cfgclk_en <= '1';
                        adc_timeout_counter <= (others => '0');
                        WRITE_SD <= '1';
                        if (CFG_DONE = '1') then
                            if (cfg_function = "01") then    -- If ADC mode and ADCs ok to read
                                state <= adc_readback;       -- go to readback adc
                             else                            -- If not ADC mode
                                state <= cfg_reset;          -- we are done, go to cfg_reset
                            end if;
                        else
						    state <= cfg_active;
                        end if;                            
                    when adc_readback =>                     -- reading telemetry data
                        fsm_state <= "110";
                        cfgclk_en <= '1';
                        WRITE_SD <= '0';                     -- stop writing serial data
                                                             -- but keep cfgclk on until ADC_DATA_READY or timeout
                        if (ADC_DATA_READY = '1') then
                            state <= cfg_reset;               
                        else
                            state <= adc_readback;
                        end if;
                    when cfg_reset =>
                        reset_state_counter <= reset_state_counter + 1;
                        fsm_state <= "111";
                        RESET_CFG <= '1';                   -- reset to cfg_ser_out and cfg_ser_in
                        if (reset_state_counter < "111111") then
                            state <= cfg_reset;             -- go back to idle state after
                        else                                -- four CFGCLK cycles
                            state <= idle;
                        end if;
                end case;
            end if;
        end if;
    end process;

end Behavioral;
