------------------------------------------------------------------------
-- PROJECT          :   MONSOON
-- BOARD            :   IR CLOCK&BIAS BOARD
-- DEVICE ID        :   VIRTEX FPGA (U116)
-- PARENT MODULE    :   TOP LEVEL
-- MODULE           :   SERIAL CONFIGURATION DATA GENERATOR
-- VERSION          :   1.4 
------------------------------------------------------------------------
-- FILE NAME        :   cfg_ser_out.vhd
-- AUTHOR           :   G. Rahmer
-- DATE             :   14 JANUARY 2003
-- COMPANY          :   NOAO
-- PLATFORM         :   Xilinx ISE Webpack 5.1
------------------------------------------------------------------------
-- DESCRIPTION:   
-- This module provides the serial data stream for the configuration
-- devices on the CFG CPLD. This stream is assembled out of parallel 
-- data provided by either the backplane (DAC, OE, MUX) or the 
-- cfg_reg_file (ADC), and of address information coming from the 
-- backplane.
-- There are four different types of streams to be generated, according
-- to CFG_SEL :
--          00->DAC
--          01->ADC
--          10->OE
--          11->MUX
--        
-- The different stream formats (18 bits) are as follows:
--      DAC:
--          |S| DAC_CHIP(5) | DAC_CHAN(4) |   DAC_Data(8)  |
--      ADC:
--          |S| ADC_CHIP(5) | fill(4) |  Control_Byte (8)  |
--      OE:
--          |S|  OE (4)   |           fill (13)            |
--      MUX (input from the data bus):
--          |L|~ENP2|0|EN3(~EN2)|B3|B2|B1|B0|0|~ENP1|0|EN1(~EN0)|A3|A2|A1|A0|
--      MUX (output stream):
--          |S|L|~ENP2|EN3(~EN2)|B3|B2|B1|B0|~ENP1|EN1(~EN0)|A3|A2|A1|A0|fill(4)|
--
--    S is a "Start bit" ("0")
--    "Fill" bits are 1's, except for ADC, which are 0's.
--
--    Control_Byte format is | 1 | ADC_CHAN(3) | ADC_Data(4) |
--    For the details of the ADC_Data, see MAX1270 datasheet.
--
--    OE bits are: | BIAS_EN0 | BIAS_EN1 | FBIAS_EN | CLK_EN |
--
--    L is a "led on/off" bit.
--
-- DAC_CHIP and DAC_CHAN fields are taken from the backplane address bus.
-- DAC_Data is taken from the backplane data bus.
--
-- ADC_CHIP and ADC_CHAN are taken from the backplane address bus.
-- ADC_Data is taken from the local data bus (coming from "cfg_reg_file").
--
-- OE and MUX data words are taken from the backplane data bus (no address).
--
-- Data stream generation is triggered by WRITE_SD.
-- 
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES:
--
-- The DAC chips have 12 channels each, and the DAC memory area starts at h00,
-- therefore:
--      dac_chip = ADDR_BP_L(7:0) DIV 12
--      dac_chan = ADDR_BP_L(7:0) MOD 12
-- 
-- The ADC chips have 8 channels each, and the ADC memory area starts at h70,
-- therefore:
--      adc_chip = ADDR_BP_L(7:3) - "01110"
--      adc_chan = ADDR_BP_L(2:0)
--      
------------------------------------------------------------------------
-- REVISION HISTORY:
-- DATE     AUTHOR  VERSION     MODIFICATIONS
-- 21/Apr/03   GR      1.1      Code completed
-- 28/Apr/03   GR      1.1      Fixed some bugs
-- 01/Jul/03   GR      1.2      Fixed dac/oe/mux_data taken erroneously from DATA_LOCAL_IN,
--                              cleaned up the code.
-- 04/Dec/03   GR      1.3      Changed the definition for the MUX serial stream 
-- 04/May/04   GR      1.4      Re-mapped the addresses for the clock level voltages
--                              in process DAC_addr_coding 
--
------------------------------------------------------------------------
-- PORT DEFINITIONS:
--
-- CFGCLK : clock signal for the CFG CPLD, generated by "cfg_control"
-- RESET : reset from the backplane
-- WRITE_SD : start of serial stream generation
-- ADDR_BP_L(7:0) : backplane address latched in cfg_control
-- DATA_LOCAL_IN(15:0) : local data bus from cfg_reg_file
-- DATA_BP_L_IN(15:0) : backplane data latched in cfg_control
-- CFG_SEL(1:0) : function select from cfg_control
-- CFG_DONE : goes high at the end of the tx of the stream. Tells the 
--            cfg_control to stop CFGCLK (except for the ADCs)
-- SER_DATA_OUT : serial data output to the CFG CPLD
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity cfg_ser_out is
    Port ( CFGCLK : in std_logic;
           RESET : in std_logic;
           WRITE_SD : in std_logic;
           ADDR_BP_L : in std_logic_vector(7 downto 0);
           DATA_LOCAL_IN : in std_logic_vector(15 downto 0);
           DATA_BP_L_IN : in std_logic_vector(15 downto 0);
           CFG_SEL : in std_logic_vector(1 downto 0);
           CFG_DONE : out std_logic;
           SER_DATA_OUT : out std_logic);
end cfg_ser_out;

architecture Behavioral of cfg_ser_out is

    type statetype is (idle, load, tx_data);
    signal state : statetype;
    signal OpReg : std_logic_vector(17 downto 0);
    signal srl_counter : std_logic_vector(4 downto 0);

    signal dac_chip : std_logic_vector(4 downto 0);
    signal dac_chan : std_logic_vector(7 downto 0);
    signal dac_data : std_logic_vector(7 downto 0);
    signal dac_OpReg : std_logic_vector(17 downto 0);

    signal adc_chip : std_logic_vector(4 downto 0);
    signal adc_chan : std_logic_vector(2 downto 0);
    signal adc_data : std_logic_vector(3 downto 0);
    signal adc_OpReg : std_logic_vector(17 downto 0);

    signal oe_data  : std_logic_vector(3 downto 0);
    signal oe_OpReg : std_logic_vector(17 downto 0);

    signal mux_data : std_logic_vector(15 downto 0);
    signal mux_OpReg : std_logic_vector(17 downto 0);

-- Auxiliary signals for re-mapping the clock level addresses
    signal bias_addr  : std_logic_vector(7 downto 0);
    signal clock_low_addr : std_logic_vector(7 downto 0);
    signal clock_high_addr : std_logic_vector(7 downto 0);
    signal remap_addr : std_logic_vector(7 downto 0);

begin

-- Bias cfg addresses
    bias_addr <= ADDR_BP_L;
-- Clock level cfg addresses        
    clock_low_addr <= ADDR_BP_L + ADDR_BP_L - x"24";
    clock_high_addr <= ADDR_BP_L + ADDR_BP_L - x"63";
                     
    remap_addr <= bias_addr when ((ADDR_BP_L >= x"00") and (ADDR_BP_L <= x"23")) else
                  clock_low_addr when ((ADDR_BP_L >= x"24") and (ADDR_BP_L <= x"43")) else
                  clock_high_addr when ((ADDR_BP_L >= x"44") and (ADDR_BP_L <= x"63")) else
                  x"00";

    DAC_addr_coding: process (remap_addr)
    begin
        if (remap_addr < x"0C") then
            dac_chip <= "00000";
            dac_chan <= remap_addr;
        elsif (remap_addr < x"18") then
            dac_chip <= "00001";
            dac_chan <= remap_addr - x"0C";
        elsif (remap_addr < x"24") then
            dac_chip <= "00010";
            dac_chan <= remap_addr - x"18";
        elsif (remap_addr < x"30") then
            dac_chip <= "00011";
            dac_chan <= remap_addr - x"24";
        elsif (remap_addr < x"3C") then
            dac_chip <= "00100";
            dac_chan <= remap_addr - x"30";
        elsif (remap_addr < x"48") then
            dac_chip <= "00101";
            dac_chan <= remap_addr - x"3C";
        elsif (remap_addr < x"54") then
            dac_chip <= "00110";
            dac_chan <= remap_addr - x"48";
        elsif (remap_addr < x"60") then
            dac_chip <= "00111";
            dac_chan <= remap_addr - x"54";
        elsif (remap_addr < x"6C") then
            dac_chip <= "01000";
            dac_chan <= remap_addr - x"60";
        else
            dac_chip <= (others => '0');
            dac_chan <= (others => '0');
        end if;            
    end process;

    dac_data <= DATA_BP_L_IN(7 downto 0);
    dac_OpReg <= '0' & dac_chip & dac_chan(3 downto 0) & dac_data;

    adc_chip <= (ADDR_BP_L(7 downto 3) - "01110");
    adc_chan <= ADDR_BP_L(2 downto 0);
    adc_data <= DATA_LOCAL_IN(3 downto 0);
    adc_OpReg <= '0' & adc_chip & "00001" & adc_chan & adc_data;

    oe_data <= DATA_BP_L_IN(3 downto 0);
    oe_OpReg <= '0' & oe_data & "1111111111111";

    mux_data <= DATA_BP_L_IN(15 downto 0);
    mux_OpReg <= '0' & mux_data(15 downto 14) & mux_data(12 downto 8) & mux_data(6) & mux_data(4 downto 0) & "1111";

    FSM: process (CFGCLK)
    begin
        if (rising_edge(CFGCLK)) then
            if (RESET = '1') then
    			state <= idle;
                srl_counter <= (others => '0'); 
		        SER_DATA_OUT <= '1';
                CFG_DONE <= '0';
            else
				case state is
					when idle =>
						SER_DATA_OUT <= '1';
                        CFG_DONE <= '0';
                        srl_counter <= (others => '0');
						if (WRITE_SD = '1') then
						    state <= load;
						else
						    state <= idle;	 	
						end if;
					when load =>
                        case CFG_SEL is
                            when "00" =>
                                OpReg <= dac_OpReg;
                            when "01" =>
                                OpReg <= adc_OpReg;
                            when "10" =>
                                OpReg <= oe_OpReg;
                            when "11" =>
                                OpReg <= mux_OpReg;
                            when others =>
                                OpReg <= "111111111111111111";
                        end case;
                        state <= tx_data;
                    when tx_data =>
                        if (srl_counter < "10011") then         -- Size of serial data stream: 18 bits + 1 "stop bit"
                            state <= tx_data;
                            srl_counter <= srl_counter + '1';
                            SER_DATA_OUT <= OpReg(17);
                            OpReg <= OpReg(16 downto 0) & '1';
                        else
                            CFG_DONE <= '1';
                            state <= idle;
                        end if;
                    when others =>
                        state <= idle;
                end case;
            end if;
        end if;
    end process;

end Behavioral;
