------------------------------------------------------------------------
-- PROJECT          :   MONSOON
-- BOARD            :   IR CLOCK&BIAS BOARD
-- DEVICE ID        :   VIRTEX FPGA (U116)
-- PARENT MODULE    :   TOP LEVEL
-- MODULE           :   SEQUENCER (CLOCK/FAST_BIAS) CONTROL MODULE
-- VERSION          :   4.90 
------------------------------------------------------------------------
-- FILE NAME        :   seq_control_v41.vhd
-- AUTHOR           :   Peter Moore
-- DATE             :   9 DECEMBER 2006
-- COMPANY          :   NOAO
-- PLATFORM         :   Xilinx ISE Webpack 7.1
------------------------------------------------------------------------
-- DESCRIPTION:   
-- This module provides the control signals for the SEQ CPLD. This
-- is a write-only access, and the backplane data goes directly to the 
-- CPLD through CPLD_DATA(31:0).
--
-- The SEQ_SEL bits are set according to the type of data, which is 
-- defined by the corresponding address:
--      00 -> clock pattern (addr = "0x0000")
--      01 -> fast biases 0-3 (addr = "0x0001")
--      10 -> fast biases 4-7 (addr = "0x0002")
--      11 -> reset (no address)
-- 
-- The Fast Bias data is previously stored in banks (0-7). Each bank
-- corresponds to one FBias channel, and contains sixteen 8-bit values.
-- The banks are located in contiguous memory addresses starting from 0x0040.
--
-- To trigger a Fast Bias update, a write command is issued with the 
-- address of the corresponding group (0-3 or 4-7), the data being the
-- desired index number (within the bank).
--
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES:
-- The fast bias DACs (AD9709) require two consecutive clocks to
-- generate the voltage. One to latch the data at the input stage, and
-- the next to pass it to the internal DAC. The value at the input must
-- be stable during both edges.
--
-- Only one group of Fast Biases can be updated at the same time.
-- All channels in each group are updated simultaneously. It is not
-- possible to update only one channel independently.
--
-- Note that a backplane generated reset condition (at RST) is coded in
-- the SEQ_SEL code "11", which is also used in the CFG CPLD.
--
------------------------------------------------------------------------
-- REVISION HISTORY:
-- DATE       AUTHOR  VERSION   MODIFICATIONS
-- 25/May/03    GR      1.2     Added registration of data and select
-- 16/Feb/04    GR      1.3     Implemented automatic double write for the fast bias data
--                              (required by the parallel DAC AD9709)
-- 10/May/04    GR      1.4     Added bank store of Fast Bias data
-- 24/Sep/04    GR      4.1     Changed the bank addressing from the top level point of view 
--                              to eight banks of 16 values each (one per channel) instead of
--                              16 banks of eight channels each.
-- 12/07/2004   PCM     4.2     ADJUSTED RAM WIDTH TO REDUCE COMPLEXITY AND RESOURCE USAGE.
--	01/21/2006   PCM     4.90    Added functionality for OTA detector DOP emulation.
-- 12/09/2006   PCM     4.90    For some reason, the bank addressing was screwed up .. Rehashed
--                              the fbias addressing scheme to work properly !
------------------------------------------------------------------------
-- PORT DEFINITIONS:
--
-- CLK : board clock
-- RESET : reset signal from the backplane interface
-- WR : write signal from the backplane interface
-- ADDR(4:0) : address from the backplane interface
-- CONTROL(15:0) : control register word (not used)
-- SEQ_CLK_OUT : clock for the SEQ CPLD (output)
-- SEQ_SEL(1:0) : function select for the SEQ CPLD (output)
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity seq_control_v490 is
    Port ( SYSCLK       : in std_logic;
           RESET        : in std_logic;
           WRITE_32     : in std_logic;
           WRITE_16     : in std_logic;
           ADDR         : in std_logic_vector(7 downto 0);
           DATA_IN      : in std_logic_vector(31 downto 0);
           CS_FB_DATA   : in std_logic;
           CFG_DATA_OUT : out std_logic_vector(7 downto 0);
           SEQ_CLK_OUT  : out std_logic;
           SEQ_SEL      : out std_logic_vector(1 downto 0);
           SEQ_DATA_OUT : out std_logic_vector(31 downto 0);
		   DOP_DATA_IN  : in  std_logic_vector(15 downto 0);
		   DOP_STROBE   : in  std_logic;
		   DOP_ACK      : inout std_logic;
           DOP_ENABLE   : in std_logic
          );
end seq_control_v490;

architecture Behavioral of seq_control_v490 is

    signal seqclk_en : std_logic := '0';

    signal sel_reg  : std_logic_vector(1 downto 0);
    signal data_reg : std_logic_vector(31 downto 0);
    signal clk_reg, clk_data, dop_data  : std_logic_vector(31 downto 0);

    signal fb_second_write : std_logic := '0';

    type ram_store_type is array (15 downto 0) of std_logic_vector(31 downto 0);
    signal fbiasRam0 : ram_store_type;
    signal fbiasRam1 : ram_store_type;

    signal fbiasAddr0 : std_logic_vector(3 downto 0);
    signal fbiasAddr1 : std_logic_vector(3 downto 0); 
    signal fbDatain   : std_logic_vector(31 downto 0);
    signal fbData0, fbData1  : std_logic_vector(31 downto 0);
    signal fbDat0,fbDat1,fbDat2,fbDat3  : std_logic_vector(7 downto 0);
    signal fbBnk0Enbl, fbBnk1Enbl  : std_logic;
begin

------------------------------------------------------------------------
-- FAST BIAS DATA STORE ADDRESS REMAPPING
------------------------------------------------------------------------
    fbiasAddr0 <= DATA_IN(3 downto 0) when (WRITE_32 = '1') else ADDR(3 downto 0);
    fbiasAddr1 <= DATA_IN(3 downto 0) when (WRITE_32 = '1') else ADDR(3 downto 0);

    fbBnk0Enbl  <= not ADDR(7);
    fbBnk1Enbl  <= ADDR(7);

    fbDataIn   <= (fbDat3 & fbDat2 & fbDat1 & DATA_IN(7 downto 0)) when (ADDR(5 downto 4) = "00") else
                  (fbDat3 & fbdat2 & DATA_IN(7 downto 0) & fbDat0) when (ADDR(5 downto 4) = "01") else
                  (fbDat3 & DATA_IN(7 downto 0) & fbDat1 & fbDat0) when (ADDR(5 downto 4) = "10") else
                  (DATA_IN(7 downto 0) & fbDat2 & fbDat1 & fbDat0);

------------------------------------------------------------------------
-- FAST BIAS DATA STORE WRITE/READ PROCESS
------------------------------------------------------------------------

    CfgStoreWrtProc0 : process (SYSCLK)
    begin
        if(rising_edge(SYSCLK)) then
            if (CS_FB_DATA = '1' and WRITE_16 = '1' and fbBnk0Enbl = '1') then
                fbiasRam0(conv_integer(fbiasAddr0)) <= fbDataIn;
            end if;
        end if;
    end process;

    CfgStoreWrtProc1 : process (SYSCLK)
    begin
        if(rising_edge(SYSCLK)) then
            if (CS_FB_DATA = '1' and WRITE_16 = '1' and fbBnk1Enbl = '1') then
                fbiasRam1(conv_integer(fbiasAddr1)) <= fbDataIn;
            end if;
        end if;
    end process;

    fbData0 <= fbiasRam0(conv_integer(fbiasAddr0));
    fbData1 <= fbiasRam1(conv_integer(fbiasAddr1));

    fbDat3  <= fbData0(31 downto 24) when (fbBnk0Enbl = '1') else fbData1(31 downto 24);
    fbDat2  <= fbData0(23 downto 16) when (fbBnk0Enbl = '1') else fbData1(23 downto 16);
    fbDat1  <= fbData0(15 downto 8) when (fbBnk0Enbl = '1') else fbData1(15 downto 8);
    fbDat0  <= fbData0(7 downto 0) when (fbBnk0Enbl = '1') else fbData1(7 downto 0);

    CFG_DATA_OUT <= fbDat0 when (ADDR(1 downto 0) = "00") else
                    fbDat1 when (ADDR(1 downto 0) = "01") else
                    fbDat2 when (ADDR(1 downto 0) = "10") else
                    fbDat3;
           
------------------------------------------------------------------------
-- DATA SEQUENCE OUTPUT PROCESS
------------------------------------------------------------------------
    SEQ_SEL <= sel_reg;
    SEQ_DATA_OUT <= clk_reg when (sel_reg = "00") else data_reg;
    SEQ_CLK_OUT <= (not SYSCLK) and seqclk_en;

	Signal_Register: process(SYSCLK)
	begin
        if (rising_edge(SYSCLK)) then
            if (fb_second_write = '1') then
                seqclk_en <= '1'; 
                fb_second_write <= '0';
            elsif ( RESET = '1' ) then
				DOP_ACK <= not DOP_ENABLE;	
                seqclk_en <= '0';
                sel_reg <= "11";
                fb_second_write <= '0';
            elsif (WRITE_32 = '1') then
				    DOP_ACK <= not DOP_ENABLE;	
                if (ADDR = "00000") then 
                    sel_reg <= "00";
                    if (DOP_ENABLE = '1') then
                        clk_reg <= clk_data;
                    else
                        clk_reg <= not DATA_IN;
                    end if;
                    seqclk_en <= '1';
                    fb_second_write <= '0';
                elsif (ADDR = "00001") then 
                    sel_reg <= "01";
                    data_reg <= fbData0;
                    seqclk_en <= '1';
                    fb_second_write <= '1';
                elsif (ADDR = "00010") then 
                    sel_reg <= "10";
                    data_reg <= fbData1;
                    seqclk_en <= '1';
                    fb_second_write <= '1';
                else
                    sel_reg <= "00";
                    seqclk_en <= '0';
                    fb_second_write <= '0';
                end if;
            elsif (DOP_STROBE = '1' and DOP_ACK = '0' and DOP_ENABLE = '1') then
                sel_reg <= "00";
                clk_reg <= dop_data;
                seqclk_en <= '1';
                fb_second_write <= '0';
				DOP_ACK <= '1';	
            else
                sel_reg <= "00";
                fb_second_write <= '0';
                seqclk_en <= '0';    
				DOP_ACK <= not DOP_ENABLE;	
            end if;
        end if;
    end process;


    clk_data <= clk_reg(31) & clk_reg(30) & not DATA_IN(29) & not DATA_IN(28) & clk_reg(27) & clk_reg(26) & not DATA_IN(25) & not DATA_IN(24) &
                clk_reg(23) & clk_reg(22) & not DATA_IN(21) & not DATA_IN(20) & clk_reg(19) & clk_reg(18) & not DATA_IN(17) & not DATA_IN(16) &
                clk_reg(15) & clk_reg(14) & not DATA_IN(13) & not DATA_IN(12) & clk_reg(11) & clk_reg(10) & not DATA_IN(9) & not DATA_IN(8) &
                clk_reg(7) & clk_reg(6) & not DATA_IN(5) & not DATA_IN(4) & clk_reg(3) & clk_reg(2) & not DATA_IN(1) & not DATA_IN(0);

    dop_data <= not DOP_DATA_IN(2) & not DOP_DATA_IN(3) & clk_reg(29) & clk_reg(28) & not DOP_DATA_IN(1) & not DOP_DATA_IN(0) & clk_reg(25) & clk_reg(24) &
                not DOP_DATA_IN(6) & not DOP_DATA_IN(7) & clk_reg(21) & clk_reg(20) & not DOP_DATA_IN(5) & not DOP_DATA_IN(4) & clk_reg(17) & clk_reg(16) &
                not DOP_DATA_IN(14) & not DOP_DATA_IN(15) & clk_reg(13) & clk_reg(12) & not DOP_DATA_IN(13) & not DOP_DATA_IN(12) & clk_reg(9) & clk_reg(8) &
                not DOP_DATA_IN(10) & not DOP_DATA_IN(11) & clk_reg(5) & clk_reg(4) & not DOP_DATA_IN(9) & not DOP_DATA_IN(8) & clk_reg(1) & clk_reg(0);

end Behavioral;
